/*
 * Detector.cpp
 *
 *  Created on: 12/giu/2010
 *      Author: fisch
 */

#define CV_NO_BACKWARD_COMPATIBILITY

#include "cv.h"
#include "highgui.h"

#include <iostream>
#include <cstdio>

#include "FaceDetector.h"

using namespace std;
using namespace cv;

FaceDetector::FaceDetector()
{
}

FaceDetector::FaceDetector( String frontalCascadeName,
							String profileCascadeName,
							String rEyeCascadeName,
							String lEyeCascadeName,
							String mouthCascadeName )
{
	if( !frontalCascade.load( frontalCascadeName ) )
		cerr << "ERROR: Could not load classifier cascade" << endl;
	if( !profileCascade.load( profileCascadeName ) )
		cerr << "ERROR: Could not load classifier cascade" << endl;
	if( !rEyeCascade.load( rEyeCascadeName ) )
		cerr << "WARNING: Could not load classifier cascade for nested objects" << endl;
	if( !lEyeCascade.load( lEyeCascadeName ) )
		cerr << "WARNING: Could not load classifier cascade for nested objects" << endl;
	if( !mouthCascade.load( mouthCascadeName ) )
		cerr << "WARNING: Could not load classifier cascade for nested objects" << endl;

	roll = 0.0; pitch = 0.0; yaw = 0.0;
	initStartMatrix( roll, pitch, yaw );

	modelFound = false;
	faceFound = false;
	resizeROI = 0.4;
	flags = 0;
	count = 0;
	lost = 0;

	pyramid = cvCreateImage( cvSize(640,480), 8, 1 );
	prev_pyramid = cvCreateImage( cvSize(640,480), 8, 1 );

	double initGlPositMatrix[16] = { 1.0, 0.0, 0.0, 0.0,
									 0.0, 1.0, 0.0, 0.0,
									 0.0, 0.0, 1.0, 0.0,
									 0.0, 0.0, 0.0, 1.0 };
	for( int i = 0; i < 16; i++ )
		glPositMatrix[i] = initGlPositMatrix[i];
}

FaceDetector::FaceDetector( String cascadeName, String nestedCascadeName )
{
	String frontalCascadeName =
	"/usr/local/share/opencv/haarcascades/haarcascade_frontalface_alt2.xml";
	String profileCascadeName =
	"/usr/local/share/opencv/haarcascades/haarcascade_profileface.xml";
	String rEyeCascadeName =
	"/usr/local/share/opencv/haarcascades/haarcascade_mcs_righteye.xml";
	String lEyeCascadeName =
	"/usr/local/share/opencv/haarcascades/haarcascade_mcs_lefteye.xml";
	String mouthCascadeName =
	"/usr/local/share/opencv/haarcascades/haarcascade_mcs_mouth.xml";

	if( !frontalCascade.load( frontalCascadeName ) )
		cerr << "ERROR: Could not load classifier cascade" << endl;
	if( !profileCascade.load( profileCascadeName ) )
		cerr << "ERROR: Could not load classifier cascade" << endl;
	if( !rEyeCascade.load( rEyeCascadeName ) )
		cerr << "WARNING: Could not load classifier cascade for nested objects" << endl;
	if( !lEyeCascade.load( lEyeCascadeName ) )
		cerr << "WARNING: Could not load classifier cascade for nested objects" << endl;
	if( !mouthCascade.load( mouthCascadeName ) )
		cerr << "WARNING: Could not load classifier cascade for nested objects" << endl;

	roll = 0.0; pitch = 0.0; yaw = 0.0;
	initStartMatrix( roll, pitch, yaw );

	modelFound = false;
	faceFound = false;
	resizeROI = 0.4;
	flags = 0;
	count = 0;
	lost = 0;

	pyramid = cvCreateImage( cvSize(640,480), 8, 1 );
	prev_pyramid = cvCreateImage( cvSize(640,480), 8, 1 );

	double initGlPositMatrix[16] = { 1.0, 0.0, 0.0, 0.0,
									 0.0, 1.0, 0.0, 0.0,
									 0.0, 0.0, 1.0, 0.0,
									 0.0, 0.0, 0.0, 1.0 };
	for( int i = 0; i < 16; i++ )
		glPositMatrix[i] = initGlPositMatrix[i];
}



bool FaceDetector::drawModel()
{
	return modelFound;
}

Point FaceDetector::getPoint()
{
	return center;
}

int FaceDetector::getRadius()
{
	return radius;
}

void FaceDetector::getReferenceCoordinate( float& x, float& y, float& z )
{
	x = referenceCoordinateX;
	y = referenceCoordinateY;
	z = referenceCoordinateZ;
}

void FaceDetector::getGLOrientation( float& roll, float& pitch, float& yaw )
{
	double myGlPositMatrix[16];
	getGlPositMatrix( myGlPositMatrix );
	roll  = atan2(-myGlPositMatrix[4], myGlPositMatrix[0] );
	pitch = atan2( myGlPositMatrix[9],-myGlPositMatrix[10] );
	yaw   = atan2( myGlPositMatrix[8], myGlPositMatrix[0]/cos(roll) );
}

void FaceDetector::getGlPositMatrix( double myGlPositMatrix[16] )
{
	myGlPositMatrix[0] = glPositMatrix[0]*startMatrix[0]+glPositMatrix[4]*startMatrix[1]+glPositMatrix[8]*startMatrix[2]+glPositMatrix[12]*startMatrix[3];
	myGlPositMatrix[4] = glPositMatrix[0]*startMatrix[4]+glPositMatrix[4]*startMatrix[5]+glPositMatrix[8]*startMatrix[6]+glPositMatrix[12]*startMatrix[7];
	myGlPositMatrix[8] = glPositMatrix[0]*startMatrix[8]+glPositMatrix[4]*startMatrix[9]+glPositMatrix[8]*startMatrix[10]+glPositMatrix[12]*startMatrix[11];
	myGlPositMatrix[12] = glPositMatrix[0]*startMatrix[12]+glPositMatrix[4]*startMatrix[13]+glPositMatrix[8]*startMatrix[14]+glPositMatrix[12]*startMatrix[15];

	myGlPositMatrix[1] = glPositMatrix[1]*startMatrix[0]+glPositMatrix[5]*startMatrix[1]+glPositMatrix[9]*startMatrix[2]+glPositMatrix[13]*startMatrix[3];
	myGlPositMatrix[5] = glPositMatrix[1]*startMatrix[4]+glPositMatrix[5]*startMatrix[5]+glPositMatrix[9]*startMatrix[6]+glPositMatrix[13]*startMatrix[7];
	myGlPositMatrix[9] = glPositMatrix[1]*startMatrix[8]+glPositMatrix[5]*startMatrix[9]+glPositMatrix[9]*startMatrix[10]+glPositMatrix[13]*startMatrix[11];
	myGlPositMatrix[13] = glPositMatrix[1]*startMatrix[12]+glPositMatrix[5]*startMatrix[13]+glPositMatrix[9]*startMatrix[14]+glPositMatrix[13]*startMatrix[15];

	myGlPositMatrix[2] = glPositMatrix[2]*startMatrix[0]+glPositMatrix[6]*startMatrix[1]+glPositMatrix[10]*startMatrix[2]+glPositMatrix[14]*startMatrix[3];
	myGlPositMatrix[6] = glPositMatrix[2]*startMatrix[4]+glPositMatrix[6]*startMatrix[5]+glPositMatrix[10]*startMatrix[6]+glPositMatrix[14]*startMatrix[7];
	myGlPositMatrix[10] = glPositMatrix[2]*startMatrix[8]+glPositMatrix[6]*startMatrix[9]+glPositMatrix[10]*startMatrix[10]+glPositMatrix[14]*startMatrix[11];
	myGlPositMatrix[14] = glPositMatrix[2]*startMatrix[12]+glPositMatrix[6]*startMatrix[13]+glPositMatrix[10]*startMatrix[14]+glPositMatrix[14]*startMatrix[15];

	myGlPositMatrix[3] = glPositMatrix[3]*startMatrix[0]+glPositMatrix[7]*startMatrix[1]+glPositMatrix[11]*startMatrix[2]+glPositMatrix[15]*startMatrix[3];
	myGlPositMatrix[7] = glPositMatrix[3]*startMatrix[4]+glPositMatrix[7]*startMatrix[5]+glPositMatrix[11]*startMatrix[6]+glPositMatrix[15]*startMatrix[7];
	myGlPositMatrix[11] = glPositMatrix[3]*startMatrix[8]+glPositMatrix[7]*startMatrix[9]+glPositMatrix[11]*startMatrix[10]+glPositMatrix[15]*startMatrix[11];
	myGlPositMatrix[15] = glPositMatrix[3]*startMatrix[12]+glPositMatrix[7]*startMatrix[13]+glPositMatrix[11]*startMatrix[14]+glPositMatrix[15]*startMatrix[15];
}



bool FaceDetector::faceDetect( unsigned char* img_pointer, double scale, double angle )
{
	IplImage* newImage;
	newImage = cvCreateImage(cvSize(640,480), IPL_DEPTH_8U, 3);
	newImage->imageData = (char*)img_pointer;
	Mat img( newImage );

	if( !lost || lost > MAX_LOST )
	{
		lost = 0;
		if( !faceFound )
			detect( img, 2 );
		else
			update( img, 0 );
	} else
		correct( img );

	draw( img );
	cvReleaseImage( &newImage );
	cvWaitKey(2);

	if( !lost )
		return faceFound;
	else
		return true;
}

bool FaceDetector::detect( Mat& img, double scale )
{
	if( count%2 == 0 )
		return detectFace( img, scale, FRONTALE );
	if( count%4 == 1 )
		return detectFace( img, scale, PROFILE_R );
	if( count%4 == 3 )
		return detectFace( img, scale, PROFILE_L );
	return false;
}

bool FaceDetector::detectFace( Mat& img, double scale, int mode )
{
	count++;
    CascadeClassifier cascade;
    modelFound = false;

    if( mode == FRONTALE )
    	cascade = frontalCascade;
    if( mode == PROFILE_R || mode == PROFILE_L )
    	cascade = profileCascade;

    cvtColor( img, gray, CV_BGR2GRAY );


	if( !cascade.empty() )
	{
		if( mode == PROFILE_L )
			flip( gray, gray, 1 );

		Rect face = detectBiggestHaar( gray, cascade, Size(40,40), cvRound(img.cols/scale) );

		if( mode == PROFILE_L )
			flip( gray, gray, 1 );

		if( face.width )
		{
			faceFound = true;

			if( mode == PROFILE_L )
				face.x = img.cols - face.x - face.width;

			roll = 0.0; pitch = 0.0; yaw = 0.0;
			modelFound = findOrientation( gray, face, roll, pitch, yaw, mode );
			initStartMatrix( roll, pitch, yaw );

			if( faceFound )
			{
				flags = 0;
				count = 0;
				insertNewPoints( gray, face, mode );
				create3DModel( face );
				initHeadWidth = face.width;
				radius = cvRound(face.width/2);
				center.x = face.x + radius;
				center.y = face.y + radius;
			}
		}
	}

	return faceFound;
}

bool FaceDetector::update( Mat& img, double dt )
{
	count++;
	int changed = 0, notStatus = 0;
	float totErr = 0.0;
	IplImage* swap_temp;

	prev_points = points;
	prev_gray = gray.clone();
	cvtColor( img, gray, CV_BGR2GRAY );
	CV_SWAP( prev_pyramid, pyramid, swap_temp );

	ransacDeleted.clear();

	if( !prev_points.empty() )
		trackOptFlow( prev_gray, gray, prev_pyramid, pyramid,
					  points, prev_points,
					  flags, changed, notStatus, totErr );

	if( count < 2 || ( changed > (POINT_PERC/100)*(int)points.size() && !notStatus ) )//&& totErr < 250 ) )
	{
		CvMatr32f rotation_matrix = new float[9];
		CvVect32f translation_vector = new float[3];

		if( points.size() >= MIN_POINTS_POSIT )
		{
			getPositMatrix( rotation_matrix, translation_vector );
			updateGLPositMatrix( rotation_matrix, translation_vector );
		}
	} else
	{
		points = prev_points;
		gray = prev_gray.clone();
	}

	float new_roll = 0.0, new_pitch = 0.0, new_yaw = 0.0;
	if( modelFound )
		getGLOrientation( new_roll, new_pitch, new_yaw );

	bool old = false;
	if( (count > MAX_FACE) || (points.size() < MIN_POINTS) || (abs(new_roll) > MAX_ANGLE) || (abs(new_pitch) > MAX_ANGLE) )
		lost = 1;
	if( count < 2 || changed < 0.8*points.size() )
		lost = 0;

	if( count < 2)
		initHeadWidth = initHeadWidth*glPositMatrix[14];
	else
		if( !lost )
		{
			updatePoint();
			updateRadius();
		}

	if( (count > OLD_FACE) && (abs(new_roll) < PI/18) && (abs(new_pitch) < PI/9) && (abs(new_yaw) < PI/9) )
		old = true;

	roll = new_roll; pitch = new_pitch; yaw = new_yaw;
	if( !modelFound || lost || old )
		faceFound = correct( img );

	return faceFound;
}

bool FaceDetector::correct( Mat& img )
{
	Rect search;
	Rect detected;
	Mat gray, grayROI;
	int newWidth;

	cvtColor( img, gray, CV_BGR2GRAY );

    Point p1( center.x-radius, center.y-radius );
    Point p2( center.x+radius, center.y+radius );

	if( lost )
	{
		lost++;
		newWidth = 100;
		resizeSearch( Rect(p1,p2), search, Size( gray.cols, gray.rows ), 1 );
	} else
	{
		newWidth = 60;
		resizeSearch( Rect(p1,p2), search, Size( gray.cols, gray.rows ), 0.5 );
	}

	grayROI = gray(search);

	int mode = FRONTALE;
	detected = detectBiggestHaar( grayROI, frontalCascade, Size(20, 20), newWidth );
	if( !detected.width && lost )
	{
		mode = PROFILE_R;
		detected = detectBiggestHaar( grayROI, profileCascade, Size(20, 20), newWidth );
	}
	if( !detected.width && lost )
	{
		mode = PROFILE_L;
		flip( grayROI, grayROI, 1 );
		detected = detectBiggestHaar( grayROI, profileCascade, Size(20, 20), newWidth );
		flip( grayROI, grayROI, 1 );
		detected.x = grayROI.cols - detected.x - detected.width;
	}

	if( detected.width )
	{
		faceFound = true;

		detected = detected + Point( search.x, search.y );

		roll = 0.0; pitch = 0.0; yaw = 0.0;
		modelFound = findOrientation( gray, detected, roll, pitch, yaw, mode );
		initStartMatrix( roll, pitch, yaw );

		if( ( modelFound || lost ) && faceFound )
		{
			lost = 0;
			flags = 0;
			count = 0;
			insertNewPoints( gray, detected, mode );
			create3DModel( detected );
			initHeadWidth = detected.width;
			radius = cvRound(detected.width/2);
			center.x = detected.x + radius;
			center.y = detected.y + radius;
		}

		return true;
	}

	if( lost )
		return false;

	return true;
}

void FaceDetector::draw( Mat& img )
{
	if( !isnan(glPositMatrix[14]) && radius >= 0 )
		circle( img, center, radius, cvScalar(0,0,255), 1, 8, 0 );

	if( modelFound )
	{
		Point vett( radius*sin(yaw)*cos(roll)+center.x, radius*(-sin(pitch))*sin(roll)+center.y );
		line( img, center, vett, cvScalar(0,0,255), 3, 8, 0 );
	}

    if( (!prev_points.empty()) && (!points.empty()) )
    {
	vector<Point2f>::const_iterator p = prev_points.begin();
	for( vector<Point2f>::const_iterator r = points.begin(); r != points.end(); r++, p++ )
	{
		line( img, Point(cvRound(p->x),cvRound(p->y)), Point(cvRound(r->x),cvRound(r->y)), CV_RGB(200,0,0), 1 );
		circle( img, Point(cvRound(r->x),cvRound(r->y)), 3, CV_RGB(0,0,200), 0, 8, 0 );
	}
    }
    if( !ransacDeleted.empty() )
    	for( vector<Point2f>::const_iterator r = ransacDeleted.begin(); r != ransacDeleted.end(); r++ )
			circle( img, Point(cvRound(r->x),cvRound(r->y)), 3, CV_RGB(0,200,200), 2, 8, 0 );

	circle( img, Point(cvRound(xc),cvRound(yc)), 3, CV_RGB(0,255,0), -1, 8, 0 );
	circle( img, Point(cvRound(xr),cvRound(yr)), 3, CV_RGB(0,255,0), -1, 8, 0 );
	circle( img, Point(cvRound(xl),cvRound(yl)), 3, CV_RGB(0,255,0), -1, 8, 0 );
	circle( img, Point(cvRound(x1),cvRound(y1)), 3, CV_RGB(0,255,0), -1, 8, 0 );
	circle( img, Point(cvRound(x2),cvRound(y2)), 3, CV_RGB(0,255,0), -1, 8, 0 );

    imshow( "result", img );
}



void FaceDetector::updatePoint()
{
	double myGlPositMatrix[16];
	getGlPositMatrix( myGlPositMatrix );

	float a[] = { myGlPositMatrix[0], myGlPositMatrix[4], myGlPositMatrix[8],  myGlPositMatrix[12],
				  myGlPositMatrix[1], myGlPositMatrix[5], myGlPositMatrix[9],  myGlPositMatrix[13],
				  myGlPositMatrix[2], myGlPositMatrix[6], myGlPositMatrix[10], myGlPositMatrix[14],
				  0,  				  0,  				  0,  				   1				  };

	float projectionMatrix[16];
	double farPlane = 10000.0;
	double nearPlane = 1.0;
	double width = 640;
	double height = 480;
	double focalLength = EHCIFOCUS;

	projectionMatrix[0] = 2*focalLength/width;
	projectionMatrix[1] = 0.0;
	projectionMatrix[2] = 0.0;
	projectionMatrix[3] = 0.0;

	projectionMatrix[4] = 0.0;
	projectionMatrix[5] = 2*focalLength/ height;
	projectionMatrix[6] = 0.0;
	projectionMatrix[7] = 0.0;

	projectionMatrix[8] = 0;
	projectionMatrix[9] = 0;
	projectionMatrix[10] =-( farPlane+nearPlane ) / ( farPlane - nearPlane );
	projectionMatrix[11] =-2.0 * farPlane * nearPlane / ( farPlane - nearPlane );

	projectionMatrix[12] = 0.0;
	projectionMatrix[13] = 0.0;
	projectionMatrix[14] =-1.0;
	projectionMatrix[15] = 0.0;

	CvMat Ma = cvMat( 4, 4, CV_32FC1, a );
	CvMat Mp = cvMat( 4, 4, CV_32FC1, projectionMatrix );

	float deltaX = referenceCoordinateX - 0.5;
	float deltaY = referenceCoordinateY + 0.5;
	float deltaZ = referenceCoordinateZ + 0.5;

	float origem[4][1];
	origem[0][0] = -deltaX;
	origem[1][0] = -0.132375-deltaY;
	origem[2][0] = 0.14104-deltaZ;
	origem[3][0] = 1.0;

	CvMat Mpoint = cvMat( 4, 1, CV_32FC1, &origem );

	CvMat* Mr1 =  cvCreateMat( 4, 1, CV_32FC1 );

	float look[] = {-1.0, 0.0, 0.0, 0.0,
					 0.0, 1.0, 0.0, 0.0,
					 0.0, 0.0,-1.0, 0.0,
					 0.0, 0.0, 0.0, 1.0 };
	CvMat Mlook = cvMat( 4, 4, CV_32FC1, look );

	cvMatMul( &Ma, &Mpoint, Mr1 );
	cvMatMul( &Mp, Mr1, Mr1 );
	cvMatMul( &Mlook, Mr1, Mr1 );

	center.x = cvmGet(Mr1,0,0)/cvmGet(Mr1,3,0)*320 + 320;
	center.y = 240 - cvmGet(Mr1,1,0)/cvmGet(Mr1,3,0)*240;

	if( center.x < 0 || center.x > 640 || center.y < 0 || center.y > 480 )
	    	faceFound = false;

	cvReleaseMat(&Mr1);
}

void FaceDetector::updateRadius()
{
	radius = cvRound((initHeadWidth/glPositMatrix[14])/2);
}

void FaceDetector::setReferenceCoordinate( float x, float y, float z )
{
	referenceCoordinateX = x;
	referenceCoordinateY = y;
	referenceCoordinateZ = z;
}

void FaceDetector::initStartMatrix( float roll, float pitch, float yaw )
{
	startMatrix[0] = cos(roll)*cos(yaw);
	startMatrix[1] = sin(roll)*cos(pitch)+cos(roll)*sin(yaw)*sin(pitch);
	startMatrix[2] = sin(roll)*sin(pitch)-cos(roll)*sin(yaw)*cos(pitch);
	startMatrix[3] = 0.0;

	startMatrix[4] =-sin(roll)*cos(yaw);
	startMatrix[5] = cos(roll)*cos(pitch)-sin(roll)*sin(yaw)*sin(pitch);
	startMatrix[6] = cos(roll)*sin(pitch)+sin(roll)*sin(yaw)*cos(pitch);
	startMatrix[7] = 0.0;

	startMatrix[8] = sin(yaw);
	startMatrix[9] =-cos(yaw)*sin(pitch);
	startMatrix[10] = cos(yaw)*cos(pitch);
	startMatrix[11] = 0.0;

	startMatrix[12] = 0.0;
	startMatrix[13] = 0.0;
	startMatrix[14] = 0.0;
	startMatrix[15] = 1.0; //homogeneous
}

void FaceDetector::updateGLPositMatrix( CvMatr32f rotation_matrix,
									CvVect32f translation_vector )
{
	glPositMatrix[0] = rotation_matrix[0];
	glPositMatrix[1] = rotation_matrix[3];
	glPositMatrix[2] = rotation_matrix[6];
	glPositMatrix[3] = 0.0;

	glPositMatrix[4] = rotation_matrix[1];
	glPositMatrix[5] = rotation_matrix[4];
	glPositMatrix[6] = rotation_matrix[7];
	glPositMatrix[7] = 0.0;

	glPositMatrix[8] = rotation_matrix[2];
	glPositMatrix[9] = rotation_matrix[5];
	glPositMatrix[10] = rotation_matrix[8];
	glPositMatrix[11] = 0.0;

	glPositMatrix[12] = translation_vector[0];
	glPositMatrix[13] = translation_vector[1];
	glPositMatrix[14] = translation_vector[2];
	glPositMatrix[15] = 1.0; //homogeneous
}



void FaceDetector::insertNewPoints( Mat grayOr, Rect face, int mode )
{
	int width = cvRound(0.7*face.width);
	int height = cvRound(0.7*face.height);
	int headX = face.x;
	int headY = face.y+cvRound(0.15*face.height);

	if( mode == FRONTALE )
		headX += cvRound(0.15*face.width);
	if( mode == PROFILE_R )
		headX += cvRound(0.3*face.width);
	if( mode == FRONTALE && !modelFound )
	{
		width = cvRound(0.5*face.width);
		headX += cvRound(0.25*face.width);
	}

	Mat gray = grayOr.clone();
	equalizeHist( gray, gray );

	Mat mask( gray.size(), CV_8UC1 );
	mask.setTo( Scalar(GC_BGD) );
	Mat maskRect = mask( Range(headY, headY+height), Range(headX, headX+width) );
	maskRect.setTo( Scalar(GC_PR_FGD) );

	double quality = 0.005;
	double minDistance = 0.005*width;
	int numPoints = 200;

    goodFeaturesToTrack( gray, points, numPoints, quality, minDistance, mask, 3, false, 0.04);
}

void FaceDetector::create3DModel( Rect face )
{
	//3D model building
	modelPoints.clear();
	for( int i = 0; i < (int)points.size(); i++ )
	{
		int px = cvRound(points[i].x) - face.x ;
		int py = cvRound(points[i].y) - face.y ;

		float fx,fy,fz;
		get3dModelCoordinates( px, py, fx, fy, fz, face.width, face.height );

		modelPoints.push_back( Point3f( fx * EHCIMODELSCALE,
									   -fy * EHCIMODELSCALE,
										fz * EHCIMODELSCALE ) );

		if( i == 0 )
			setReferenceCoordinate( fx * EHCIMODELSCALE,
								   -fy * EHCIMODELSCALE,
									fz * EHCIMODELSCALE );
	}
}

/**
 * Converts to the sinusoidal model
 * Points should be relative to upper left head corner
 */
void FaceDetector::get3dModelCoordinates( int px, int py, float& fx, float& fy, float& fz, int headWidth, int headHeight )
{
	fx = (1.66667 * px/(1.0*headWidth)) - 0.33333;
	fy = (1.66667 * py/(1.0*headHeight)) - 0.33333;
	fz = .5*sin(fx*3.141593);
}



bool FaceDetector::findOrientation( Mat grayOr, Rect face, float& roll, float& pitch, float& yaw, int pose )
{
	Mat gray = grayOr.clone();
	equalizeHist( gray, gray );

	if( pose == FRONTALE )
	{
	    Rect resize = face + Size(-cvRound(face.width*resizeROI),-cvRound(face.height*resizeROI));
	    Rect lEye = detectBiggestHaar( gray( resize ),
											lEyeCascade, Size(18, 12), 40 );
	    Rect rEye = detectBiggestHaar( gray( resize + Point(cvRound(face.width*resizeROI),0) ),
											rEyeCascade, Size(18, 12), 40 );
	    rEye.x = rEye.x + cvRound(face.width*resizeROI);
	    Rect mouth = detectBiggestHaar( gray( resize + Point(cvRound(face.width*resizeROI/2),cvRound(face.height*resizeROI)) ),
											mouthCascade, Size(18, 15), 100 );
	    mouth = mouth + Point(cvRound(face.width*resizeROI/2),cvRound(face.height*resizeROI));

		if( rEye.width && lEye.width && mouth.width )
		{
			Mat g1 = gray.clone();
			Mat g2 = gray.clone();
			Point2f rCenter = findPupilCenter( g1, rEye+Point(face.x,face.y));
			Point2f lCenter = findPupilCenter( g2, lEye+Point(face.x,face.y));
			Point2f mCenter = findPhiltrumCenter( gray, mouth+Point(face.x,face.y));

			xc = mCenter.x; yc = mCenter.y;
			xl = face.x+lEye.x+(double)lEye.width/2; yl = face.y+lEye.y+(double)lEye.height/2;
			xr = face.x+rEye.x+(double)rEye.width/2; yr = face.y+rEye.y+(double)rEye.height/2;

			x1 = (xr-xl)/2+xl, y1 = (yr-yl)/2+yl;
			double mpa = (yr-yl)/(xr-xl), mpe = (xl-xr)/(yr-yl);
			double qpa = yc-mpa*xc, qpe = y1-mpe*x1;
			x2 = (qpa-qpe)/(mpe-mpa), y2 = mpa*x2 + qpa;

			double dista = sqrt((xc-x2)*(xc-x2)+(yc-y2)*(yc-y2));
			double distb = sqrt((xr-xl)*(xr-xl)+(yr-yl)*(yr-yl));

			if( xc > x2 ) dista = -dista;
			yaw = -atan2( 3*dista, distb );

			roll = -atan2( (double)(rCenter.y-lCenter.y), (double)(rCenter.x-lCenter.x) );

			return true;
		}
		if( !rEye.width && !lEye.width )
			faceFound = false;
	}

	if( pose == PROFILE_R )
	{
		Rect resize = face + Size(0,-cvRound(face.height*resizeROI));
	    Rect rEye = detectBiggestHaar( gray( resize ),
											rEyeCascade, Size(18, 12), 80 );
		if( !rEye.width )
			faceFound = false;
	}

	if( pose == PROFILE_L )
	{
		Rect resize = face + Size(0,-cvRound(face.height*resizeROI));
	    Rect lEye = detectBiggestHaar( gray( resize ),
											lEyeCascade, Size(18, 12), 80 );
		if( !lEye.width )
			faceFound = false;
	}

	return false;
}

Point2f FaceDetector::findCentroid( Mat gray, Rect search )
{
	Mat grayROI = gray( search );

	threshold( grayROI, grayROI, 100, 255, THRESH_BINARY_INV );
	Moments m = moments( gray( search ), false );
	return Point2f( m.m10/m.m00 + search.x, m.m01/m.m00 + search.y);
}

Point2f FaceDetector::findPupilCenter( Mat gray, Rect eye )
{
	Rect search( cvRound(eye.x+0.2*eye.width),cvRound(eye.y+0.2*eye.height),cvRound(0.6*eye.width),cvRound(0.6*eye.height) );
	return findCentroid( gray, search );
}

Point2f FaceDetector::findPhiltrumCenter( Mat gray, Rect mouth )
{
	return Point2f( (float)mouth.x+(float)mouth.width/2, (float)mouth.y-(float)mouth.height/5 );
}



void FaceDetector::resizeFaceRect( Rect src, Rect& dest, Size dimMax )
{
	dest = src;
	if(dest.x < 0)
	{
		dest.width = dest.width + dest.x;
		dest.x = 0;
	}
	if(dest.y < 0)
	{
		dest.height = dest.height + dest.y;
		dest.y = 0;
	}
	if(dest.width < 0)
		dest.width = dimMax.width;
	if(dest.height < 0)
		dest.height = dimMax.height;
	if(dest.x + dest.width > dimMax.width)
		dest.width = dimMax.width - dest.x;
	if(dest.y + dest.height > dimMax.height)
		dest.height = dimMax.height - dest.y;
}

void FaceDetector::resizeSearch( Rect src, Rect& dest , Size dimMax, float factor )
{
	dest = src;

	dest = dest - Point(cvRound((factor/2)*dest.width), cvRound((factor/2)*dest.height));
	dest = dest + Size(cvRound(factor*dest.width), cvRound(factor*dest.height));

	resizeFaceRect( dest, dest, dimMax );
}



void FaceDetector::getPositMatrix( CvMatr32f rotation_matrix, CvVect32f translation_vector )
{
	vector<Point2f> imagePoints;
	vector<int> outliers;

	for( vector<Point2f>::const_iterator r = points.begin(); r != points.end(); r++ )
		imagePoints.push_back( Point2f( cvRound(r->x)-320, cvRound(r->y)-240 ) );

	if( !modelPoints.empty() )
		ransac( imagePoints, modelPoints, rotation_matrix, translation_vector, outliers );

	if( !outliers.empty() )
		correctPoints( outliers );
}

void FaceDetector::correctPoints( vector<int> outliers )
{
	int tCount = 0;
	int target = outliers[tCount];

	vector<Point3f> updatedModelPoints;
	vector<Point2f> updatedPrevPoints;
	vector<Point2f> updatedPoints;

	for( int i = 0; i < (int)modelPoints.size(); i++ )
	{
		if( i == target )
		{
			ransacDeleted.push_back(points[i]);

			if( tCount < (int)outliers.size()-1 )
				target = outliers[++tCount];//update target
		}
		else
		{
			updatedModelPoints.push_back(modelPoints[i]);
			updatedPrevPoints.push_back(prev_points[i]);
			updatedPoints.push_back(points[i]);
		}
	}

	modelPoints.clear();
	modelPoints = updatedModelPoints;
	prev_points.clear();
	prev_points = updatedPrevPoints;
	points.clear();
	points = updatedPoints;
}



Rect detectBiggestHaar( Mat img, CascadeClassifier cascade, Size cascadeSize, int newWidth )
{
	Rect result;
	Mat imgCopy;
	Mat newImg;
	vector<Rect> detections;

	imgCopy = img.clone();
	equalizeHist( imgCopy, imgCopy );

	double scale = (double)img.cols / (double)newWidth;
	int newHeight = cvRound( (double)img.rows / scale );
	resize( imgCopy, newImg, Size( newWidth, newHeight ), 0, 0, INTER_LINEAR );

	if( !cascade.empty() )
		cascade.detectMultiScale( newImg, detections, 1.1, 1,
								  CV_HAAR_FIND_BIGGEST_OBJECT, cascadeSize );

	if( !detections.empty() )
	{
		result = detections.front();
		result = Rect( cvRound(result.x*scale), cvRound(result.y*scale),
					   cvRound(result.width*scale), cvRound(result.height*scale) );
	}

	return result;
}

void trackOptFlow( Mat prev_gray, Mat gray, IplImage* prev_pyramid, IplImage* pyramid,
				   vector<Point2f>& points, vector<Point2f>& prev_points,
				   int& flags, int& changed, int& notStatus, float& totErr )
{
	vector<uchar> status;
	vector<float> err;

	IplImage d_prev_grey = prev_gray;
	IplImage d_grey = gray;

	IplImage* prev_grey = &d_prev_grey;
	IplImage* grey = &d_grey;

	CvPoint2D32f *old_points, *old_prev_points;
	float* old_err;
	char* old_status;
	CvPoint2D32f pt; pt.x = 0.0; pt.y = 0.0;

	old_prev_points = (CvPoint2D32f*)cvAlloc(points.size()*sizeof(pt));
	old_points = (CvPoint2D32f*)cvAlloc(points.size()*sizeof(pt));
	old_err = (float*)cvAlloc(points.size()*sizeof(float));
	old_status = (char*)cvAlloc(points.size());

	for( int i = 0; i < (int)points.size(); i++ )
	{
		old_points[i] = points[i];
		old_prev_points[i] = prev_points[i];
	}

	cvCalcOpticalFlowPyrLK( prev_grey, grey, prev_pyramid, pyramid,
							old_prev_points, old_points, prev_points.size(),
							cvSize(10,10), 3, old_status, old_err,
							cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS,20,0.03), flags );

	flags |= CV_LKFLOW_PYR_A_READY;

	for( int i = 0; i < (int)points.size(); i++ )
	{
		points[i] = old_points[i];
		prev_points[i] = old_prev_points[i];
		status.push_back(old_status[i]);
		err.push_back(old_err[i]);
	}

	vector<Point2f> valid_points;
	vector<Point2f> prev_valid_points;
	vector<float>::const_iterator e = err.begin();
	vector<uchar>::const_iterator s = status.begin();
	vector<Point2f>::const_iterator p = prev_points.begin();
	for( vector<Point2f>::const_iterator r = points.begin(); r != points.end(); r++, p++, s++, e++ )
	{
		if( !*s )
		{
			notStatus++;
			continue;
		}
		valid_points.push_back( *r );
		prev_valid_points.push_back( *p );

		double dx = r->x - p->x;
		double dy = r->y - p->y;
		if( dx*dx + dy*dy >= 2 )
			changed++;

		totErr += *e;
	}

	points = valid_points;
	prev_points = prev_valid_points;

	totErr = totErr/(float)points.size();
}

/**
 * This function is implemented as in 'Multiple View Geometry in Computer Vision'
 * 1 - Randomly select samples from the 2d tracked set
 * 2 - Use POSIT from this set
 * 3 - Use distance to filter outliers
 * 4 - Keep doing until some threshold of inliers size is found
 */
int ransac( vector<Point2f> imagePoints, vector<Point3f> objectPoints,
			CvMatr32f rotation_matrix, CvVect32f translation_vector,
			vector<int>& outliers )
{
	vector<int> bestSet;
	CvMatr32f rotationTestMatrix = new float[9];
	CvVect32f translationTestVector = new float[3];
	int bestSize = 0;

	for( int i = 0; i < (int)imagePoints.size(); i++ )
		bestSet.push_back(i);

	for( int k = 0; k < RANSAC_ITERATIONS; k++ )
	{
		//only choose first 4 points:
		vector<int> randomPoints = getRandomSet( imagePoints.size(),
												 RANSAC_SAMPLES );

		getMatrixUsingPosit( imagePoints, objectPoints, randomPoints,
							 rotationTestMatrix, translationTestVector );

		vector<int> inliers;
		plot2dModel( rotationTestMatrix, translationTestVector, objectPoints, imagePoints, inliers );

		if( (int)inliers.size() > bestSize )
		{
			bestSize = inliers.size();
			bestSet.clear();
			outliers.clear();

			int dCount = 0;
			for( int i = 0; i < (int)inliers.size(); i++ )
			{
				while( dCount != inliers[i] )
				{
					outliers.push_back(dCount);
					dCount++;
				}

				bestSet.push_back(inliers[i]);
				dCount++;
			}
			while( dCount < (int)imagePoints.size() )
			{
				outliers.push_back(dCount);
				dCount++;
			}
		}
	}

	delete [] rotationTestMatrix;
	delete [] translationTestVector;

	//generate rot and trans matrixes with best inliers
	//now use the bestSet to retrieve the rotation and translation matrixes
	return getMatrixUsingPosit( imagePoints, objectPoints, bestSet, rotation_matrix, translation_vector );
}

int getMatrixUsingPosit( vector<Point2f> imagePoints,
						 vector<Point3f> objectPoints,
						 vector<int> selectedPoints,
						 CvMatr32f rotation_matrix,
						 CvVect32f translation_vector )
{
	vector<CvPoint3D32f> chosenPoints;
	vector<CvPoint2D32f> chosenImagePoints;
	//do not change the first one, because it's the reference
	chosenPoints.push_back(objectPoints[0]);
	chosenImagePoints.push_back(imagePoints[0]);

	for( int i = 1; i < (int)selectedPoints.size(); i++ )
	{
		int index = selectedPoints[i];
		if( index < (int)objectPoints.size() )
		{
			chosenPoints.push_back(objectPoints[index]);
			chosenImagePoints.push_back(imagePoints[index]);
		}
	}

	if( selectedPoints.size() >= 4 )
	{
		CvPOSITObject *positObject = cvCreatePOSITObject( &chosenPoints[0], static_cast<int>(chosenPoints.size()) );
		//set posit termination criteria: 1000 max iterations, convergence epsilon 1.0e-5

		CvTermCriteria criteria = cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS, 1000, 1.0e-5 );
		cvPOSIT( positObject, &chosenImagePoints[0], EHCIFOCUS, criteria, rotation_matrix, translation_vector );

		cvReleasePOSITObject (&positObject);
	}
	else
	{
		printf("Error: no posit: number of points %d\n",chosenPoints.size());
		return 1;//error
	}

	return 0;
}

/**
 * Returns the inliers indexes in the inliers vector
 * returns the sum of the distances
 */
double plot2dModel( CvMatr32f rotation_matrix, CvVect32f translation_vector,
					vector<Point3f> objectPoints, vector<Point2f> imagePoints,
					vector<int>& inliers )
{
	if( objectPoints.empty() )
		return 100000000.0;

	float a[] = { rotation_matrix[0], rotation_matrix[1], rotation_matrix[2], translation_vector[0],
				  rotation_matrix[3], rotation_matrix[4], rotation_matrix[5], translation_vector[1],
				  rotation_matrix[6], rotation_matrix[7], rotation_matrix[8], translation_vector[2],
								   0,				   0,				   0,					  1 };

	float projectionMatrix[16];
	double farPlane = 10000.0;
	double nearPlane = 1.0;
	double width = 640;
	double height = 480;
	double focalLength = EHCIFOCUS;

	projectionMatrix[0] = 2*focalLength/width;
	projectionMatrix[1] = 0.0;
	projectionMatrix[2] = 0.0;
	projectionMatrix[3] = 0.0;

	projectionMatrix[4] = 0.0;
	projectionMatrix[5] = 2*focalLength/ height;
	projectionMatrix[6] = 0.0;
	projectionMatrix[7] = 0.0;

	projectionMatrix[8] = 0;
	projectionMatrix[9] = 0;
	projectionMatrix[10] =-( farPlane+nearPlane ) / ( farPlane - nearPlane );
	projectionMatrix[11] =-2.0 * farPlane * nearPlane / ( farPlane - nearPlane );

	projectionMatrix[12] = 0.0;
	projectionMatrix[13] = 0.0;
	projectionMatrix[14] =-1.0;
	projectionMatrix[15] = 0.0;

	CvMat Ma = cvMat( 4, 4, CV_32FC1, a );

	CvMat Mp = cvMat( 4, 4, CV_32FC1, projectionMatrix );

	int numPoints = objectPoints.size();

	int w;
	CvMat Mpoints[500];
	float pontos[500][4];

	for( w = 0; w < numPoints; w++ )
	{
		//fazendo tambem a mudanca de coordenada para o ponto 0
		pontos[w][0] =  objectPoints[w].x - objectPoints[0].x;
		pontos[w][1] =  objectPoints[w].y - objectPoints[0].y;
		pontos[w][2] =  objectPoints[w].z - objectPoints[0].z;
		pontos[w][3] =  1.0;

		Mpoints[w] = cvMat( 4, 1, CV_32FC1, &pontos[w] );
	}

	float origem[4];
	origem[0] = 0 - objectPoints[0].x;
	origem[1] = 0 - objectPoints[0].y;
	origem[2] = 0 - objectPoints[0].z;
	origem[3] = 1.0;

	Mpoints[w] = cvMat( 4, 1, CV_32FC1, &origem );

	float origem1[4];
	origem1[0] = 1.0 - objectPoints[0].x;
	origem1[1] =-1.0 - objectPoints[0].y;
	origem1[2] = 0.0 - objectPoints[0].z;
	origem1[3] = 1.0;

	Mpoints[w+1] = cvMat( 4, 1, CV_32FC1, &origem1 );

	CvMat* Mr1 =  cvCreateMat( 4, 1, CV_32FC1 );

	float look[] = {-1.0, 0.0, 0.0, 0.0,
					 0.0, 1.0, 0.0, 0.0,
					 0.0, 0.0,-1.0, 0.0,
					 0.0, 0.0, 0.0, 1.0 };
	CvMat Mlook = cvMat( 4, 4, CV_32FC1, look );

	double distance = 0;
	double threshold = getDistanceThreshold( Ma, Mp, Mpoints, Mlook,
											 numPoints, imagePoints );

	if( !isnan(threshold) )
	{
		for( w = 0; w < numPoints; w++ )
		{
			cvMatMul( &Ma, &Mpoints[w], Mr1 );
			cvMatMul( &Mp, Mr1, Mr1 );

			//fazer o look at para o posit
			cvMatMul( &Mlook, Mr1, Mr1 );

			double xWinPosition = cvmGet(Mr1,0,0)/cvmGet(Mr1,3,0)*320;
			double yWinPosition = cvmGet(Mr1,1,0)/cvmGet(Mr1,3,0)*240;
			double dx = xWinPosition - imagePoints[w].x ;
			double dy = yWinPosition - (-imagePoints[w].y);

			double pointDistance = sqrt(dx*dx+dy*dy);

			distance += pointDistance;

			if( pointDistance <= RANSAC_DISTANCE_THRESHOLD )
				inliers.push_back(w);
		}
	}

	cvReleaseMat(&Mr1);
	return distance;
}

/**
 * Returns the threshold based in a 95% inlier probability,
 * in case a gaussian distribuition for the error is used as hypothesis
 * according to Multiple View Geometry in Computer Vision
 */
double getDistanceThreshold( CvMat Ma, CvMat Mp, CvMat* Mpoints, CvMat Mlook,
							 int numPoints, vector<Point2f> imagePoints )
{
	vector<double> distanceErrors;
	CvMat* Mr1 = cvCreateMat(4,1,CV_32FC1);

	double errorSum = 0;

	for( int w = 0; w < numPoints; w++ )
	{
		cvMatMul( &Ma, &Mpoints[w], Mr1 );
		cvMatMul( &Mp, Mr1, Mr1 );

		//fazer o look at para o posit
		cvMatMul( &Mlook, Mr1, Mr1 );

		double xWinPosition = cvmGet(Mr1,0,0)/cvmGet(Mr1,3,0)*320;
		double yWinPosition = cvmGet(Mr1,1,0)/cvmGet(Mr1,3,0)*240;
		double dx = xWinPosition - imagePoints[w].x ;
		double dy = yWinPosition - (-imagePoints[w].y );//estao com sinais trocados
		double pointDistance = sqrt(dx*dx+dy*dy);
		errorSum += pointDistance;
		distanceErrors.push_back(pointDistance);
	}

	double deviations = 0;
	double meanError = errorSum/numPoints;
	for( int i = 0; i < numPoints; i++ )
		deviations += (distanceErrors[i]-meanError)*(distanceErrors[i]-meanError);

	double sigma = sqrt(deviations/numPoints);
	cvReleaseMat(&Mr1);
	return sqrt(5.99 * sigma*sigma);
}

vector<int> getRandomSet( int sampleSize, int setSize )
{
	if( sampleSize < setSize )
	{
		printf("Set size too small in RANSAC\n");
		exit(0);
	}

	vector<int> lista;
	for( int i = 0; i < sampleSize; i++ )
		lista.push_back(i);

	vector<int> randomSet;
	while( (int)randomSet.size() < setSize )
	{
		int pos = rand()%lista.size();
		randomSet.push_back(lista[pos]);
		lista.erase(lista.begin()+pos);
	}

	return randomSet;
}
